"""Partial tests (pytest) for the gpxo module."""

import numpy as np
import gpxo
from gpxo import compass, closest_pt


def test_loadtrack():
    """Loading/smoothing/velocity calculation from position/distance."""
    track = gpxo.Track('ExampleTrack.gpx')
    track.smooth(n=51)
    assert round(track.data['velocity (km/h)'].iloc[4]) == 14


def test_loadtrack_notime():
    """Loading in situation where GPX does not have time info."""
    track = gpxo.Track('ExampleTrack_NoTime.gpx')
    assert round(track.data['compass (°)'].iloc[3]) == 92


def test_compass_single():
    """Test compass calculation with individual (lat, long) tuples."""
    assert compass((0, 0), (45, 0)) == 0.0
    assert compass((45, 0), (0, 0)) == 180.0
    assert compass((0, 0), (0, 30)) == 90.0
    assert compass((0, 30), (0, -30)) == 270.0


def test_compass_multiple():
    """Test compass calculation with tuples of iterables (lat_list, long_list)."""
    lat1 = (0, 45, 0, 0)
    long1 = (0, 0, 0, 30)
    lat2 = (45, 0, 0, 0)
    long2 = (0, 0, 30, -30)
    pt1 = (lat1, long1)
    pt2 = (lat2, long2)
    assert all(compass(pt1, pt2) == (0, 180, 90, 270))


def test_closest_pt():
    """Test find index of closest point in trajectory to specified pt."""
    lats = [45.011, 45.012, 45.013, 45.014, 45.015, 45.016, 45.017]
    longs = [5.883, 5.886, 5.887, 5.889, 5.891, 5.893, 5.895]
    traj = np.array((lats, longs))
    pt = (45.0133, 5.888)
    i = closest_pt(pt, traj)
    assert i == 2



"""General tools for gpx data processing based on gpxpy."""


import numpy as np
from vincenty import vincenty


def closest_pt(pt, trajectory):
    """Finds closest pt to pt (lat, long) in trajectory (lats, longs).

    Parameters
    ----------
    pt is a tuple (lat, long)
    trajectory can be:
    - a tuple (lats, longs) where lats is an iterable of floats (and longs also)
    - a (2 * N) numpy array where N is the length of the trajectory
    - any other structure equivalent in terms of unpacking a, b = trajectory
    """
    lats, longs = trajectory
    ds = [vincenty((x, y), pt) for (x, y) in zip(lats, longs)]
    return np.argmin(ds)


def compass(pt1, pt2):
    """
    Calculate the compass bearing between two points.

    Adapted from https://gist.github.com/jeromer/2005586 (public domain) for
    use with numpy arrays.

    The formula used is the following:
        θ = arctan2(sin(Δlong).cos(lat2),
                  cos(lat1).sin(lat2) − sin(lat1).cos(lat2).cos(Δlong))

    Parameters
    ----------
    pt1, pt2: positions (latitude/longitude in °) of first/second point

    Each pt in (pt1, pt2) can be:
    - a tuple (lat, long) where lat/long are floats or iterables of floats
      (e.g. tuples, lists, arrays)
    - a (2 * N) numpy array where 2 corresponds to (lat, long) and N the number
      of successive points to calculate.
    - any structure equivalent to the structures above in terms of unpacking.

    Output
    ------
    Compass bearing in decimal degrees (°)

    Examples
    --------

    With individual points ---------------------------------------------------

    >>> compass((0, 0), (45, 0))
    0.0

    >>> compass((45, 0), (0, 0))
    180.0

    >>> compass((0, 0), (0, 30))
    90.0

    >>> compass((0, 30), (0, -30))
    270.0

    With multiple points -----------------------------------------------------

    >>> lat1 = (0, 45, 0, 0)

    >>> long1 = (0, 0, 0, 30)

    >>> lat2 = (45, 0, 0, 0)

    >>> long2 = (0, 0, 30, -30)

    >>> pt1 = (lat1, long1)

    >>> pt2 = (lat2, long2)

    >>> compass(pt1, pt2)
    array([  0., 180.,  90., 270.])
    """
    lat1, long1 = np.radians(pt1)
    lat2, long2 = np.radians(pt2)

    d_long = long2 - long1

    x = np.sin(d_long) * np.cos(lat2)
    y = np.cos(lat1) * np.sin(lat2) - (np.sin(lat1) * np.cos(lat2) * np.cos(d_long))

    initial_bearing = np.arctan2(x, y)

    # Now we have the initial bearing but np.arctan2 return values
    # from -180° to + 180° which is not what we want for a compass bearing
    # The solution is to normalize the initial bearing as shown below
    initial_bearing = np.degrees(initial_bearing)
    compass_bearing = (initial_bearing + 360) % 360

    return compass_bearing


def smooth(x, n=5, window='hanning'):
    """Smooth 1-d data with a window of requested size and type.

    Simplified version of numbo.smooth with smaller default window size (n)
    (see https://cameleon.univ-lyon1.fr/ovincent/numbo)

    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.

    INPUT
    -----
    - `x`: the input signal
    - `n`: the dimension of the smoothing window; should be an odd integer
    - `window`: the type of window ('flat', 'hanning', 'hamming', 'bartlett',
    'blackman'); flat window will produce a moving average smoothing.

    OUTPUT
    ------
    - Smoothed signal of same length as input signal

    EXAMPLE
    -------
    npts = 100  # number of data points
    w = 7  # width of window used to smooth
    x = np.linspace(0, 7 * pi, npts)
    y = np.sin(x - 1) + 0.1 * np.random.randn(n)
    y_smooth = smooth(y, w)

    NOTES
    -----
    See also numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman,
    numpy.convolve, scipy.signal.lfilter
    """
    if x.size < n:
        raise ValueError("Input vector needs to be larger than window size.")

    if n == 1:  # no need to apply filter
        return x

    if window not in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        msg = "Only possible windows: 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'"
        raise ValueError(msg)

    xadd_left = 2 * x[0] - x[n:0:-1]
    xadd_right = 2 * x[-1] - x[-2:-n - 2:-1]
    x_expanded = np.concatenate((xadd_left, x, xadd_right))

    if window == 'flat':  # moving average
        w = np.ones(n, 'd')
    else:
        w = eval('np.' + window + '(n)')

    y = np.convolve(w / w.sum(), x_expanded, mode='valid')

    istart = int(n / 2) + 1

    return y[istart:istart + len(x)]



"""General tools for gpx data processing based on gpxpy."""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

import gpxpy
from vincenty import vincenty
import mplleaflet

from .general import smooth, closest_pt


# ========================= Misc. private functions ==========================


# Function to transform array of timedeltas to seoncds
_total_seconds = np.vectorize(lambda dt: dt.total_seconds())


# ============================ Main class (Track) ============================

class Track:

    def __init__(self, filename, track=0, segment=0):

        with open(filename, 'r') as gpx_file:
            gpx = gpxpy.parse(gpx_file)

        pts = gpx.tracks[track].segments[segment].points

        self.latitude = np.array([pt.latitude for pt in pts])
        self.longitude = np.array([pt.longitude for pt in pts])
        self.elevation = np.array([pt.elevation for pt in pts])
        self.time = np.array([pt.time for pt in pts])

        # If some elevation or time data is missing, just set attribute to None

        if any(self.time == None):
            self.time = None

        if any(self.elevation == None):
            self.elevation = None

    @staticmethod
    def _distance(position1, position2):
        """Distance between two positions (latitude, longitude)."""
        return vincenty(position1, position2)

    def _resample(self, quantity, reference):
        """Resample quantities (velocity, compass) to fall back on reference

        Reference is typically time or distance."""
        # midpoints correponding to shifted quantity
        midpts = reference[:-1] + (np.diff(reference) / 2)
        # linear interpolation to fall back to initial times
        qty_resampled = np.interp(reference, midpts, quantity)
        return qty_resampled

    @property
    def seconds(self):
        if self.time is not None:
            return _total_seconds(self.time - self.time[0])

    @property
    def distance(self):
        """Travelled distance in kilometers."""

        ds = [0]

        x1s = self.latitude[:-1]
        x2s = self.latitude[1:]

        y1s = self.longitude[:-1]
        y2s = self.longitude[1:]

        for x1, x2, y1, y2 in zip(x1s, x2s, y1s, y2s):
            dd = self._distance((x1, y1), (x2, y2))
            ds.append(dd)

        return np.cumsum(ds)

    @property
    def compass(self):
        """Compass bearing in decimal degrees (°). See gpxo.compass"""
        lat1, long1 = np.radians((self.latitude[:-1], self.longitude[:-1]))
        lat2, long2 = np.radians((self.latitude[1:], self.longitude[1:]))

        d_long = long2 - long1

        x = np.sin(d_long) * np.cos(lat2)
        y = np.cos(lat1) * np.sin(lat2) - (np.sin(lat1) * np.cos(lat2) * np.cos(d_long))

        # Resample before taking arctan because if not, interpolation fails
        # when the signal fluctuates between 0 and 360° when compass is N
        x_res = self._resample(x, self.distance)
        y_res = self._resample(y, self.distance)

        initial_bearing = np.arctan2(x_res, y_res)

        # Now we have the initial bearing but np.arctan2 return values
        # from -180° to + 180° which is not what we want for a compass bearing
        # The solution is to normalize the initial bearing as shown below
        initial_bearing = np.degrees(initial_bearing)
        compass_bearing = (initial_bearing + 360) % 360

        return compass_bearing

    @property
    def velocity(self):
        """Instantaneous velocity in km/h."""
        if self.time is not None:
            dt = np.diff(self.seconds)
            dd = np.diff(self.distance)
            vs = 3600 * dd / dt
            return self._resample(vs, self.seconds)
        else:
            return None

    @property
    def data(self):
        """pd.DataFrame with all track data (time, position, velocity etc.)"""

        names = ['latitude (°)', 'longitude (°)', 'distance (km)', 'compass (°)']
        columns = [self.latitude, self.longitude, self.distance, self.compass]

        if self.time is not None:
            names += ['time', ' duration (s)', 'velocity (km/h)']
            columns += [self.time, self.seconds, self.velocity]

        if self.elevation is not None:
            names.append('elevation (m)')
            columns.append(self.elevation)

        data = pd.DataFrame(dict(zip(names, columns)))

        if self.time is not None:
            data['time'] = data['time'].dt.tz_localize(None)
            data.set_index('time', inplace=True)

        return data

    def plot(self, *args, **kwargs):
        """Plot columns of self.data (use pandas DataFrame plot arguments)."""
        return self.data.plot(*args, **kwargs)

    def smooth(self, n=5, window='hanning'):
        """Smooth position data (and subsequently distance, velocity etc.)

        Parameters
        ----------
        - n: size of moving window for smoothing
        - window: type of window (e.g. 'hanning' or 'flat', see gpxo.smooth())
        """
        self.latitude = smooth(self.latitude, n=n, window=window)
        self.longitude = smooth(self.longitude, n=n, window=window)
        self.elevation = smooth(self.elevation, n=n, window=window)

    def closest_to(self, pt):
        """Find index of point in trajectory that is closest to pt=(lat, long)."""
        return closest_pt(pt, (self.latitude, self.longitude))

    def map(self, map_type='osm', embed=False, size=(10, 10)):
        """Plot trajectory on map.

        Parameters
        ----------
        - map_type can be e.g. osm, esri_aerial, esri_worldtopo, etc. see:
        https://github.com/jwass/mplleaflet/blob/master/mplleaflet/maptiles.py

        - embed: if True, embed plot in Jupyter. If False (default), open in
        browser.

        - size: when embedded, size of the figure.
        """
        fig, ax = plt.subplots(figsize=size)
        ax.plot(self.longitude, self.latitude, '.-r')
        parameters = {'fig': fig, 'tiles': map_type}
        if embed:
            leaflet = mplleaflet.display(**parameters)
        else:
            leaflet = mplleaflet.show(**parameters)

        return leaflet
